/**
 * 约定数据格式
 * type - 事件类型
 * from - 事件发起方
 * target - 事件目标方
 * timestamp - 事件发送时间戳
 * data - 事件传递的数据
 * browserTabId - 浏览器页签ID
 */
export interface TransferData<T = unknown> extends Record<PropertyKey, unknown> {
    type: string;
    from?: string;
    target?: string;
    timestamp?: number;
    data?: T;
    browserTabId?: string;
}
export interface ApiData<T = unknown> extends TransferData {
    type: EventBusType.GetData | EventBusType.SetData;
    api?: string;
    promise?: 'resolve' | 'reject';
    params?: object;
    data?: T;
}
export declare enum EventBusType {
    GetData = "get-data",
    SetData = "set-data"
}
type EventCB<T = unknown> = (res?: TransferData<T>) => void;
type EventCBAll<T = unknown> = (type: string, res?: TransferData<T>) => void;
type ApiCb<T = unknown, U = unknown> = (from: string, params?: T) => Promise<U>;
/**
 * event-bus 实例化配置
 * name - 唯一标识
 * children - 子应用列表，用于发送 global 事件
 * workerPath - worker 静态文件地址，用于同源多标签页或同源 iframe 通信
 * devMode - 开启时，可以在控制台输出 event-bus 通信的日志
 * browserTabId - 浏览器页签ID,用于限制事件在浏览器单标签页响应
 */
export interface BusOption {
    name: string;
    children?: string[];
    workerPath?: string;
    devMode?: boolean;
    browserTabId?: string;
}
export declare class EventBus {
    private _emitter;
    private _fromName;
    private _children;
    private _apiMap;
    private _worker;
    private _isPolyfill;
    private _version;
    private _devMode;
    private _browserTabId;
    constructor(option: BusOption);
    private get _isInIframe();
    get isInIframe(): boolean;
    get version(): string;
    private _init;
    /**
     * 监听事件，同时监听 emit 和 post 的事件
     * @param type 事件类型
     * @param cb 触发时的回调函数，接收符合 TransferData 结构的 res 参数
     */
    on<T>(type: '*', cb: EventCBAll<T>): void;
    on<T>(type: string, cb: EventCB<T>): void;
    /**
     * 发送本地事件
     * @param type 事件类型
     * @param data 传递的数据
     * @param isCrossbrowserTabs 是否跨浏览器页签
     */
    emit<T = unknown>(type: string, data?: T, isCrossbrowserTabs?: boolean): void;
    private m_emit;
    /**
     * 向子应用、父应用发送消息
     * @param data 传输的数据
     * @param target 目标 iframe
     * @param byId 是否通过 id 查找 iframe，默认通过 name 查找 iframe
     * @param isCrossbrowserTabs 是否跨浏览器页签
     */
    post<T = unknown>(data: TransferData<T>, target?: 'global' | string | string[], byId?: boolean, isCrossbrowserTabs?: boolean): void;
    /**
     * 关闭事件监听
     * @param type 事件类型
     * @param cb 关闭监听的回调函数，如果不传 cb ，则关闭该事件类型下的所有函数
     */
    off<T = unknown>(type: string, cb?: EventCB<T>): void;
    /**
     * 清除所有事件监听
     */
    clearAll(): void;
    private _emitIframe;
    private _emitIframeById;
    private _emitParent;
    /**
     * 注册 api，用于 getData / getDataSelf 获取数据
     * @param api api 名称
     * @param cb 子应用调用 api 时的回调函数，cb 返回一个 Promise，resolve 的数据即为 getData 时拿到的数据
     */
    registryApi(api: string, cb: ApiCb): void;
    private _setData;
    /**
     * 调用框架提供的 api
     * @param api 框架注册的接口名
     * @param params  需要传递的参数
     * @returns 返回一个 Promise eg. .then(res => res.data)
     */
    getData<T>(api: string, params?: object): Promise<ApiData<T>>;
    /**
     * 浏览器某个标签页或 iframe 调用自己提供的 api
     * @param api 注册的接口名
     * @param params  需要传递的参数
     * @param isCrossbrowserTabs 是否跨浏览器页签
     * @returns 返回一个 Promise eg. .then(res => res.data)
     */
    getDataSelf<T>(api: string, params?: object, isCrossbrowserTabs?: boolean): Promise<ApiData<T>>;
    /**
     * 设置子应用列表
     * @param list 传入的子应用名称列表
     * @param isMerge 是否合并列表，默认替换 event-bus 实例上的 children 列表 （false），为 true 时与 children 列表进行合并。已进行去重处理。
     */
    setChildrenApp(list: string[], isMerge?: boolean): void;
    private _clearRepeat;
    private _log;
}
export default EventBus;
