/**
 * event-bus@0.4.7
 */
import mitt, { type Handler } from 'mitt'
import { VERSION, LOCALSTORAGE_LOG_KEY } from './constant'

/**
 * 约定数据格式
 * type - 事件类型
 * from - 事件发起方
 * target - 事件目标方
 * timestamp - 事件发送时间戳
 * data - 事件传递的数据
 * browserTabId - 浏览器页签ID
 */
export interface TransferData<T=unknown> extends Record<PropertyKey, unknown> {
  type: string
  from?: string
  target?: string
  timestamp?: number
  data?: T
  browserTabId?: string
}

/**
 * tabTag - 用于跨标签页通信标记，自动生成，一般不需要使用
 */
interface TabTransferData extends TransferData {
  tabTag?: string
}

export interface ApiData<T=unknown> extends TransferData {
  type: EventBusType.GetData | EventBusType.SetData
  api?: string
  promise?: 'resolve' | 'reject'
  params?: object
  data?: T
}

export enum EventBusType {
  GetData = 'get-data',
  SetData = 'set-data',
}

type EventCB<T=unknown> = (res?: TransferData<T>) => void
type EventCBAll<T=unknown> = (type: string, res?: TransferData<T>) => void // on('*') 时的回调函数
type ApiCb<T=unknown, U=unknown> = (from: string, params?: T) => Promise<U>

/**
 * event-bus 实例化配置
 * name - 唯一标识
 * children - 子应用列表，用于发送 global 事件
 * workerPath - worker 静态文件地址，用于同源多标签页或同源 iframe 通信
 * devMode - 开启时，可以在控制台输出 event-bus 通信的日志
 * browserTabId - 浏览器页签ID,用于限制事件在浏览器单标签页响应
 */
export interface BusOption {
  name: string
  children?: string[]
  workerPath?: string
  devMode?: boolean
  browserTabId?: string
}

const POLYFILL_SHARED_KEY = 'polyfill-shared'

export class EventBus {
  private _emitter = mitt()
  private _fromName!: string
  private _children: string[] = []
  private _apiMap = new Map<string, ApiCb>()
  private _worker: SharedWorker | undefined
  private _isPolyfill = false // 对于不兼容 sharedworker 的浏览器，使用 localStorage 传递消息
  private _version = VERSION // 版本信息
  private _devMode = false // 是否开启开发模式，开发模式下会输出主要日志
  private _browserTabId // 浏览器页签id

  constructor(option: BusOption) {
    if(!option.name) {
      console.error('[Error]EventBus: The name cannot be empty during initialization')
      return
    }

    if(option.devMode) {
      this._devMode = option.devMode
    }

    this._fromName = option.name
    this._log(`EventBus 版本: ${this._version}`)

    if (option.children && option.children.length) {
      this._children = this._clearRepeat(option.children)
    }

    this._browserTabId = option?.browserTabId

    if (option.workerPath) {
      const workerPath = option.workerPath.includes('http') ? option.workerPath : location.origin + option.workerPath
      try {
        if (window.SharedWorker) {
          this._worker = new SharedWorker(workerPath)
          this._worker.port.start()
          this._log(`sharedWorker 连接成功: ${workerPath}`)
        } else {
          // 处理浏览器兼容，使用 postMessage 方案，并且将 window 存在 localStorage
          this._isPolyfill = true
          window.addEventListener('storage', e => {
            if (e.key === POLYFILL_SHARED_KEY && e.newValue) {
              const data = JSON.parse(e.newValue)
              if (data.type) {
                this.m_emit(data.type, data)
              }
            }
          })
        }
      } catch (error) {
        this._worker = undefined
        console.error(error)
      }
    }

    this._init()
  }

  private get _isInIframe() {
    return window.parent !== window
  }

  get isInIframe() {
    return window.parent !== window
  }

  get version() {
    return this._version
  }

  private _init() {
    // 监听 iframe 消息
    window.addEventListener('message', e => {
      if (e && e.data && e.data.type && (!e.data.browserTabId || !this._browserTabId ||e.data.browserTabId === this._browserTabId)) {
        const data: TransferData = e.data
        // 如果接收到 get-data 的类型，拦截并处理
        if (data.type === EventBusType.GetData) {
          const getData: ApiData = e.data
          this._setData(getData)
          return
        }
        this.m_emit(data.type, data)
      }
    })

    // 为 getDataSelf 注册监听 get-data 事件
    this.on(EventBusType.GetData, (res?: TransferData) => {
      // 过滤掉没有 tab-tag 的，因为没有 tab-tag 意味着在一个 页签/iframe 内，不需要通过 getDataSelf 来通信
      if(res?.tabTag) {
        this._setData(res as ApiData)
      }
    })

    if(this._worker) {
      // 监听当前应用所有的 emit
      this.on<TabTransferData>('*', (type, data) => {
        /**
         * 过滤掉 tabTag 防止事件循环
         * 过滤掉 target 防止[子-父，父-子]发消息时重复通过 sharedWorker 通信
         */
        if (data && !data.tabTag && !data.target) {
          this._worker?.port.postMessage(data)
        }
      })
      this._worker.port.addEventListener('message', e => {
        const data: TabTransferData = e.data
        
        if (data && data.type && (!data.browserTabId || !this._browserTabId || data.browserTabId === this._browserTabId)) {
          // 添加一个标记，避免陷入事件循环
          data.tabTag = 'tab-tag'
          this._log(`sharedWorker 接收`, data)
          this.m_emit(data.type, data)
        }
      })
      // 页面关闭和刷新的时候，清除缓存的 worker
      window.addEventListener('unload', e => {
        this._worker?.port.postMessage({type: 'page-unload'})
      })
    }
  }

  /**
   * 监听事件，同时监听 emit 和 post 的事件
   * @param type 事件类型
   * @param cb 触发时的回调函数，接收符合 TransferData 结构的 res 参数
   */
  on<T>(type: '*', cb: EventCBAll<T>): void
  on<T>(type: string, cb: EventCB<T>): void
  on<T>(type: string, cb: EventCBAll<T> | EventCB<T>) {
    this._emitter.on(type, cb as Handler<unknown>)
  }

  /**
   * 发送本地事件
   * @param type 事件类型
   * @param data 传递的数据
   * @param isCrossbrowserTabs 是否跨浏览器页签
   */
  emit<T = unknown>(type: string, data?: T, isCrossbrowserTabs?:boolean) {
    const tData: TransferData = {
      type,
      data,
    }
    if (!isCrossbrowserTabs && this._browserTabId) {
      tData.browserTabId = this._browserTabId
    }

    this.m_emit(type, tData)
  }

  private m_emit(type: string, data?: TransferData) {
    if (data && !data.timestamp) {
      data.timestamp = Date.now() + Math.random()
    }
    if (data && !data.from) {
      data.from = this._fromName
    }
    this._emitter.emit(type, data)
    if (this._isPolyfill) {
      try {
        window.localStorage.setItem(POLYFILL_SHARED_KEY, JSON.stringify(data))
      } catch (error) {
        console.error('[Error]EventBus: ', error)
      }
    }
  }

  /**
   * 向子应用、父应用发送消息
   * @param data 传输的数据
   * @param target 目标 iframe
   * @param byId 是否通过 id 查找 iframe，默认通过 name 查找 iframe
   * @param isCrossbrowserTabs 是否跨浏览器页签
   */
  post<T=unknown>(data: TransferData<T>, target?: 'global' | string | string[], byId?: boolean, isCrossbrowserTabs?:boolean) {
    data.from = this._fromName
    // 如果 target 存在，则发给指定 target，不存在则发给父级
    if (target) {
      if (!isCrossbrowserTabs && this._browserTabId) {
        data.browserTabId = this._browserTabId
      }
      // 如果 target 存在，需要给数据添加一个属性： target ，为了 sharedWorker 可以过滤框架发给子应用的消息
      data.target = typeof target === 'string' ? target : target.join(',')
      // 如果 target 为 global，则遍历 children 子应用列表，给存活的子应用发送消息
      if (target === 'global') {
        this._children.forEach(v => {
          const iframes = document.getElementsByName(v) as NodeListOf<HTMLIFrameElement>
          if (iframes && iframes.length) {
            this._emitIframe(v, data)
          }
        })
      } else if (typeof target === 'string') {
        if (byId) {
          this._emitIframeById(target, data)
        } else {
          this._emitIframe(target, data)
        }
      } else if (Array.isArray(target)) {
        target.forEach(v => {
          if (byId) {
            this._emitIframeById(v, data)
          } else {
            this._emitIframe(v, data)
          }
        })
      }
    } else {
      // 否则给 data 添加一个 target 属性: parent，为了 sharedWorker 可以过滤子应用发给框架的消息
      data.target = 'parent'
      this._emitParent(data)
    }
  }

  /**
   * 关闭事件监听
   * @param type 事件类型
   * @param cb 关闭监听的回调函数，如果不传 cb ，则关闭该事件类型下的所有函数
   */
  off<T=unknown>(type: string, cb?: EventCB<T>) {
    this._emitter.off(type, cb as Handler<unknown>)
  }

  /**
   * 清除所有事件监听
   */
  clearAll() {
    this._emitter.all.clear()
  }

  // 发送事件至 iframe 子应用
  private _emitIframe(iframeName: string, data: TransferData) {
    // 通过 iframe name 查找所有 iframe
    const iframes = document.getElementsByName(iframeName) as NodeListOf<HTMLIFrameElement>
    if (iframes && iframes.length) {
      this._log(`post to iframeName: ${iframeName}[${iframes.length}]`, data)
      iframes.forEach((v) => {
        v.contentWindow?.postMessage(data, '*')
      })
    } else {
      console.error('[Error]EventBus: Please check if the iframe[name]: [' + iframeName + '] exists')
      console.log('[Log]EventBus: ', data)
    }
  }

  // 发送事件至 iframe 子应用
  private _emitIframeById(iframeId: string, data: TransferData) {
    // 通过 iframe id 查找指定 iframe
    const iframe = document.getElementById(iframeId) as HTMLIFrameElement
    if (iframe) {
      this._log(`post to iframeId: ${iframeId}`, data)
      iframe.contentWindow?.postMessage(data, '*')
    } else {
      console.error('[Error]EventBus: Please check if the iframe[id]: [' + iframeId + '] exists')
      console.log('[Log]EventBus: ', data)
    }
  }

  // 向父应用发送消息
  private _emitParent(data: TransferData) {
    if(this._isInIframe) {
      this._log(`发送至框架`, data)
      window.parent?.postMessage(data, '*')
    }
  }

  /**
   * 注册 api，用于 getData / getDataSelf 获取数据
   * @param api api 名称
   * @param cb 子应用调用 api 时的回调函数，cb 返回一个 Promise，resolve 的数据即为 getData 时拿到的数据
   */
  registryApi (api: string, cb: ApiCb) {
    this._apiMap.set(api, cb)
  }

  // 执行 api cb 并返回
  private _setData (data: ApiData) {
    if(!data.api) {
      return console.error('[Error]EventBus: API parameter error when getData: ', data)
    }
    const cb = this._apiMap.get(data.api)
    const appName = data.from // 请求来源
    const resData: ApiData = {
      type: EventBusType.SetData,
      from: this._fromName,
      timestamp: data.timestamp,
      data: null
    }
    if (appName) {
      // 如果应用名称跟当前初始化的名称一样，则通过 emit 发送结果，如果不一样，则向子应用发送结果
      if(appName === this._fromName) {
        if(cb) {
          cb(appName, data.params).then(res => {
            resData.data = res
            this.m_emit(EventBusType.SetData, resData)
          })
        }
      } else {
        if (cb) {
          cb(appName, data.params).then(res => {
            resData.data = res
            resData.promise = 'resolve'
            this.post(resData, appName)
          }).catch(error => {
            resData.data = error
            resData.promise = 'reject'
            this.post(resData, appName)
          })
        } else {
          resData.data = {
              code: 404,
              errorMsg: 'Check the API is exist in the framework : ' + data.api 
          }
          this.post(resData, appName)
        }
      }
    }
  }

  /**
   * 调用框架提供的 api
   * @param api 框架注册的接口名
   * @param params  需要传递的参数
   * @returns 返回一个 Promise eg. .then(res => res.data)
   */
  getData<T> (api: string, params?: object): Promise<ApiData<T>> {
    return new Promise((resolve, reject) => {
      const timestamp = Date.now() + Math.random()
      this._emitParent({ 
        type: EventBusType.GetData,
        from: this._fromName,
        api,
        timestamp,
        params,
      })
      const setDataCallBack: EventCB<T> = (data?: TransferData<T>) => {
        if (data && data.timestamp === timestamp) {
          this._log(`getData 到的数据`, data)
          const apiData = data as ApiData<T>
          if(!apiData.promise) {
            resolve(apiData)
          } else {
            if(apiData.promise === 'resolve') {
              resolve(apiData)
            } else {
              reject(apiData)
            }
          }
          this.off(EventBusType.SetData, setDataCallBack)
        }
      }
      this.on<T>(EventBusType.SetData, setDataCallBack)
    })
  }

  /**
   * 浏览器某个标签页或 iframe 调用自己提供的 api
   * @param api 注册的接口名
   * @param params  需要传递的参数
   * @param isCrossbrowserTabs 是否跨浏览器页签
   * @returns 返回一个 Promise eg. .then(res => res.data)
   */
  getDataSelf<T> (api: string, params?: object, isCrossbrowserTabs?:boolean): Promise<ApiData<T>> {
    return new Promise((resolve, reject) => {
      const timestamp = Date.now() + Math.random()
      let browserTabId = ''
      if (!isCrossbrowserTabs && this._browserTabId) {
        browserTabId = this._browserTabId
      }
      this.m_emit(EventBusType.GetData, {
        type: EventBusType.GetData,
        api,
        timestamp,
        params,
        browserTabId
      })
      const setDataCallBack: EventCB = (data?: TransferData) => {
        if (data && data.timestamp === timestamp) {
          this._log(`getDataSelf 到的数据`, data)
          const apiData = data as ApiData<T>
          if(!apiData.promise) {
            resolve(apiData)
          } else {
            if(apiData.promise === 'resolve') {
              resolve(apiData)
            } else {
              reject(apiData)
            }
          }
          this.off(EventBusType.SetData, setDataCallBack)
        }
      }
      this.on<T>(EventBusType.SetData, setDataCallBack)
    })
  }

  /**
   * 设置子应用列表
   * @param list 传入的子应用名称列表
   * @param isMerge 是否合并列表，默认替换 event-bus 实例上的 children 列表 （false），为 true 时与 children 列表进行合并。已进行去重处理。
   */
  setChildrenApp (list: string[], isMerge = false) {
    if (list) {
      if (isMerge) {
        this._children = this._clearRepeat([...this._children, ...list])
      } else {
        this._children = this._clearRepeat(list)
      }
    }
  }

  // 检查并消除重复应用名
  private _clearRepeat(list: string[]): string[] {
    const appMap = new Map<string, string>()
    const apps: string[] = []
    list.forEach(v => {
      appMap.set(v, v)
    })
    appMap.forEach(v => {
      apps.push(v)
    })
    return apps
  }

  private _log(msg: string, data?: TransferData) {
    try {
      const showEventBusLog = localStorage.getItem(LOCALSTORAGE_LOG_KEY)
      if (this._devMode || showEventBusLog) {
        let logMsg = `[log]EventBus: [${this._fromName}]: ${msg}`
        if (data) {
          if (data.target) {
            logMsg = `[log]EventBus: [${this._fromName}]: ${msg}: 目标: ${data.target} , 事件类型：${data.type}`
          } else {
            logMsg = `[log]EventBus: [${this._fromName}]: ${msg} , 事件类型：${data.type}`
          }
          console.log(logMsg, data)
        } else {
          console.log(logMsg)
        }
      }
    } catch (error) {
      console.error('[Error]EventBus: ', error)
    }
  }
}

export default EventBus
